/*
    Project: TCTEMP
    Version: V1.22
    File   : misc.c

    Copyright (c) 2006-2012 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <sddl.h>
#include <ctype.h>
#include <tchar.h>
#include "project.h"

#ifdef _MSC_VER
#  pragma warning( push )
#  pragma warning( disable : 4201 )
#endif
#include <shlobj.h>
#ifdef _MSC_VER
#  pragma warning( pop )
#endif

#ifndef _WIN64
typedef BOOL (WINAPI *PFISWOW64PROCESS) (HANDLE hProcess, PBOOL pfWow64Process);
static BOOL IsWow64(void)
{
    BOOL fIsWow64 = FALSE;
    PFISWOW64PROCESS pfIsWow64Process = (PFISWOW64PROCESS) GetProcAddress(
      GetModuleHandle(TEXT("KERNEL32")), "IsWow64Process");

    if (pfIsWow64Process != NULL) {
        pfIsWow64Process(GetCurrentProcess(), &fIsWow64);
    }

    return fIsWow64;
}
#endif

static BOOL ProtectFile(LPCTSTR pszFileName, LPCTSTR pszSecDesc)
{
    LPVOID lpSecurityDescriptor;

    if (!ConvertStringSecurityDescriptorToSecurityDescriptor(
      pszSecDesc, SDDL_REVISION_1, &lpSecurityDescriptor, NULL)) {
        return FALSE;
    }

    if (!SetFileSecurity(pszFileName, DACL_SECURITY_INFORMATION,
      lpSecurityDescriptor)) {
        DWORD dwErrorCode = GetLastError();
        LocalFree(lpSecurityDescriptor);
        SetLastError(dwErrorCode);
        return FALSE;
    }
    LocalFree(lpSecurityDescriptor);

    return TRUE;
}

int GetType(LPTSTR pszType)
{
    const static struct {
        LPTSTR pszName;
        int    iType;
    } Type[] = {
        { TEXT("REG_SZ"), REG_SZ },
        { TEXT("REG_MULTI_SZ"), REG_MULTI_SZ },
        { TEXT("REG_EXPAND_SZ"), REG_EXPAND_SZ },
        { TEXT("REG_DWORD"), REG_DWORD },
        { NULL, REG_NONE }
    };
    int i;
    for (i = 0; Type[i].pszName != NULL; i++) {
        if (_tcsicmp(pszType, Type[i].pszName) == 0) {
            return Type[i].iType;
        }
    }

    return REG_NONE;
}

LPTSTR GetRoot(LPTSTR pszFullKeyName, PHKEY phKey)
{
    const static struct {
        LPTSTR pszName;
        HKEY   hKey;
    } Root[] = {
        { TEXT("HKEY_CLASSES_ROOT"), HKEY_CLASSES_ROOT },
        { TEXT("HKEY_CURRENT_USER"), HKEY_CURRENT_USER },
        { TEXT("HKEY_LOCAL_MACHINE"), HKEY_LOCAL_MACHINE },
        { TEXT("HKEY_USERS"), HKEY_USERS },
        { TEXT("HKCU"), HKEY_CURRENT_USER },
        { TEXT("HKLM"), HKEY_LOCAL_MACHINE },
        { NULL, NULL }
    };
    int i;
    LPTSTR pszKeyName = _tcschr(pszFullKeyName, TEXT('\\'));

    if (pszKeyName == NULL) return NULL;
    *pszKeyName++ = TEXT('\0');

    for (i = 0; Root[i].pszName != NULL; i++) {
        if (_tcsicmp(pszFullKeyName, Root[i].pszName) == 0) {
            if (phKey) {
                *phKey = Root[i].hKey;
            }
            return pszKeyName;
        }
    }

    return NULL;
}

BOOL RegSetString(HKEY hRootKey, LPCTSTR pszKeyName, LPCTSTR pszValueName,
  DWORD dwType, LPCTSTR pszValueData)
{
    HKEY hKey;
    BOOL  fResult = FALSE;

    if (RegCreateKeyEx(hRootKey, pszKeyName, 0, NULL, 0,
      KEY_WRITE, NULL, &hKey, NULL) != ERROR_SUCCESS) {
        return FALSE;
    }
    if (RegSetValueEx(hKey, pszValueName, 0, dwType, (const BYTE*)pszValueData,
      (DWORD)(_tcslen(pszValueData)+1)*sizeof(TCHAR)) == ERROR_SUCCESS) {
        fResult = TRUE;
    }
    RegCloseKey(hKey);

    return fResult;
}

BOOL RegSetInteger(HKEY hRootKey, LPCTSTR pszKeyName, LPCTSTR pszValueName,
  DWORD dwData)
{
    HKEY hKey;
    BOOL fResult = FALSE;

    if (RegCreateKeyEx(hRootKey, pszKeyName, 0, NULL, 0,
      KEY_WRITE, NULL, &hKey, NULL) != ERROR_SUCCESS) {
        return FALSE;
    }
    if (RegSetValueEx(hKey, pszValueName, 0, REG_DWORD, (const BYTE*)&dwData,
      sizeof(DWORD)) == ERROR_SUCCESS) {
        fResult = TRUE;
    }
    RegCloseKey(hKey);

    return fResult;
}

BOOL RegRemoveKey(HKEY hRootKey, LPCTSTR pszKeyName)
{
    HKEY     hKey;
    LONG     lResult;
    TCHAR    szSubKeyName[MAX_PATH];
    DWORD    dwSize;
    FILETIME ftLastWriteTime;

    if (RegDeleteKey(hRootKey, pszKeyName) == ERROR_SUCCESS) {
        return TRUE;
    }

    lResult = RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ, &hKey);
    if (lResult != ERROR_SUCCESS) {
        if (lResult == ERROR_FILE_NOT_FOUND) {
            return TRUE;
        }
        SetLastError(lResult);
        return FALSE;
    }

    dwSize = sizeof(szSubKeyName)/sizeof(TCHAR);
    while (RegEnumKeyEx(hKey, 0, szSubKeyName, &dwSize, NULL,
      NULL, NULL, &ftLastWriteTime) == ERROR_SUCCESS) {
        if (!RegRemoveKey(hKey, szSubKeyName)) {
            DWORD dwLastError = GetLastError();
            RegCloseKey(hKey);
            SetLastError(dwLastError);
            return FALSE;
        }
        dwSize = sizeof(szSubKeyName)/sizeof(TCHAR);
    }
    RegCloseKey(hKey);

    lResult = RegDeleteKey(hRootKey, pszKeyName) == ERROR_SUCCESS;
    if (lResult == ERROR_SUCCESS) {
        return TRUE;
    }
    SetLastError(lResult);
    return FALSE;
}

BOOL RegRemoveValue(HKEY hRootKey, LPCTSTR pszKeyName,
  LPCTSTR pszValueName)
{
    HKEY hKey;
    LONG lResult;

    lResult = RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_WRITE, &hKey);
    if (lResult != ERROR_SUCCESS) {
        SetLastError(lResult);
        return FALSE;
    }

    lResult = RegDeleteValue(hKey, pszValueName);
    if (lResult != ERROR_SUCCESS) {
        RegCloseKey(hKey);
        SetLastError(lResult);
        return FALSE;
    }
    RegCloseKey(hKey);

    return TRUE;
}

LPTSTR ConvertString(LPCTSTR pszSrc, BOOL fMultiString, PDWORD pcbSize)
{
    DWORD  cbSize;
    LPTSTR pszDestBuffer;
    LPTSTR pszDest;

    cbSize = (DWORD)(_tcslen(pszSrc) + 1 + (fMultiString ? 1 : 0)) *
      sizeof(TCHAR);
    pszDestBuffer = (LPTSTR) LocalAlloc(LMEM_FIXED | LMEM_ZEROINIT, cbSize);
    if (pszDestBuffer == NULL) {
        return NULL;
    }
    pszDest = pszDestBuffer;

    while (*pszSrc) {
        if (pszSrc[0] == '&' && pszSrc[1] == '#') {
            LPTSTR psz;
            ULONG ul = _tcstoul(pszSrc+2, &psz, 10);
            if (*psz == ';' && ul < 0x10000)  {
                *pszDest++ = (TCHAR) ul;
                pszSrc = psz + 1;
                continue;
            }
        }
        *pszDest++ = *pszSrc++;
    }
    if (fMultiString) {
        pszDest += 2;
    } else {
        pszDest = pszDestBuffer + _tcslen(pszDestBuffer) + 1;
    }

    if (pcbSize) {
        *pcbSize = (DWORD)((PBYTE)pszDest - (PBYTE)pszDestBuffer);
    }

    return pszDestBuffer;
}

BOOL FileNameIsValid(LPCTSTR pszFileName)
{
    LPCTSTR psz;
    BOOL    local_filename  = FALSE;
    BOOL    remote_filename = FALSE;
    BOOL    remote_name     = FALSE;
    BOOL    server_name     = FALSE;
    BOOL    slash_mode      = FALSE;

    for (psz = pszFileName; *psz; psz++) {
        switch (*psz) {
        case TEXT('*'):
        case TEXT('?'):
        case TEXT('\"'):
        case TEXT('<'):
        case TEXT('>'):
        case TEXT('|'):
            return FALSE;
        case TEXT(':'):
            if ((psz - pszFileName) == 1) {
                local_filename = TRUE;

                if (!_istalpha(pszFileName[0])) {
                    return FALSE;
                }
                if (pszFileName[2] != TEXT('\\') &&
                  pszFileName[2] != TEXT('/')) {
                    return FALSE;
                }
            } else {
                return FALSE;
            }
            break;
        case TEXT('\\'):
            if (slash_mode) {
                if ((pszFileName - psz) == 1) {
                    server_name = TRUE;
                    continue;
                }
                return FALSE;
            }
            if (server_name) {
                if (remote_name) {
                    remote_filename = TRUE;
                } else {
                    remote_name = TRUE;
                }
            }
            slash_mode = TRUE;
            continue;
        case TEXT('/'):
            if (local_filename) {
                slash_mode = TRUE;
                continue;
            }
            return FALSE;
        }
        slash_mode = FALSE;
    }
    if (slash_mode) {
        return FALSE;
    }

    return local_filename || remote_filename;
}

BOOL VolumeNameIsValid(LPCTSTR psz)
{
    if (FileNameIsValid(psz)) {
        return TRUE;
    }
    if (_tcsncmp(psz, TEXT("\\Device\\Harddisk"), 16) != 0) {
        return FALSE;
    }
    psz += 16;
    if (!_istdigit(*psz)) {
        return FALSE;
    }
    if (psz[0] == TEXT('0') && _istdigit(psz[1])) {
        return FALSE;
    }
    while (_istdigit(*psz)) {
        psz++;
    }
    if (_tcsncmp(psz, TEXT("\\Partition"), 10) != 0) {
        return FALSE;
    }
    psz += 10;
    if (!_istdigit(*psz)) {
        return FALSE;
    }
    if (psz[0] == TEXT('0') && _istdigit(psz[1])) {
        return FALSE;
    }
    while (_istdigit(*psz)) {
        psz++;
    }

    return *psz == TEXT('\0');
}

SIZE_T MultiStringLen(LPCTSTR psz) /* return value includes terminators */
{
    LPCTSTR pszBegin = psz;

    for (;;) {
        if (psz[0] == TEXT('\0') && psz[1] == TEXT('\0')) {
            return (psz - pszBegin) + 2;
        }
        psz++;
    }
}

BOOL RemoveSingleString(LPTSTR pszBuffer, LPCTSTR pszPattern)
{
    SIZE_T cchPatternLen = _tcslen(pszPattern);
    SIZE_T cchStringLen;

    while (*pszBuffer) {
        cchStringLen = _tcslen(pszBuffer);
        if (_tcsncmp(pszBuffer, pszPattern, cchPatternLen) == 0) {
            SIZE_T cchLen = MultiStringLen(pszBuffer);
            if (pszBuffer[cchStringLen + 1] != TEXT('\0')) {
                cchStringLen++;
            }
            memmove(pszBuffer, pszBuffer + cchStringLen, sizeof(TCHAR) *
              (cchLen - cchStringLen));
            return TRUE;
        }
        pszBuffer += cchStringLen + 1;
    }

    return FALSE;
}

SIZE_T RemoveString(LPTSTR pszBuffer, LPCTSTR pszPattern)
{
    while (RemoveSingleString(pszBuffer, pszPattern)) ;
    return MultiStringLen(pszBuffer);
}

int _tmain(int argc, _TCHAR *argv[])
{
    DWORD  dwErrorCode = NO_ERROR;

    if (argc < 2) {
    syntax_error:
        _putts(
            TEXT("MISC V") TEXT(PROJECT_VERSION_TEXT) TEXT("\n\n")
            TEXT("Syntax:\n\n")
            TEXT("  MISC appentry key value data\n\n")
            TEXT("    append an entry of a REG_MULTI_SZ registry value\n\n")
            TEXT("  MISC checkstr {ascii|nospace|filename|tcfilename} string\n")
            TEXT("    returns 0 if string belongs to the specified class\n\n")
            TEXT("  MISC checkstrlen maxlen string\n")
            TEXT("    returns 0 if string length <= maxlen\n\n")
            TEXT("  MISC delentry key value start_of_entry\n\n")
            TEXT("    delete all entries of a REG_MULTI_SZ registry value\n")
            TEXT("    which start with start_of_entry\n\n")
            TEXT("  MISC delete key [value]\n\n")
            TEXT("    delete registry key or value.\n\n")
            TEXT("  MISC drivetype drive-letter\n\n")
            TEXT("    returns the drive type\n\n")
            TEXT("  MISC findistr string sub-string\n")
            TEXT("    returns 0 if sub-string is part of string (ignoring case)\n\n")
            TEXT("  MISC findstr string sub-string\n")
            TEXT("    returns 0 if sub-string is part of string\n\n")
            TEXT("  MISC isadmin\n\n")
            TEXT("    returns 0 if user is an admin.\n\n")
            TEXT("  MISC iswin64\n\n")
            TEXT("    returns 0 for a 64-bit Windows version.\n\n")
            TEXT("  MISC osversion\n\n")
            TEXT("    returns the version number of the operating system.\n\n")
            TEXT("  MISC print text\n\n")
            TEXT("    print text\n\n")
            TEXT("  MISC protectfile filename\n\n")
            TEXT("    set file DACL (System, administrators and users)\n\n")
            TEXT("  MISC protectsystemfile filename\n\n")
            TEXT("    set file DACL (System and administrators)\n\n")
            TEXT("  MISC set type key value data\n\n")
            TEXT("    type  : value type (REG_SZ, REG_MULTI_SZ, REG_EXPAND_SZ or REG_DWORD)\n")
            TEXT("    key   : registry key which must start either with\n")
            TEXT("            HKEY_LOCAL_MACHINE or HKEY_CURRENT_USER\n")
            TEXT("    value : value name\n")
            TEXT("    data  : value data\n\n")
            TEXT("    The registry value is correspondingly set.\n")
            TEXT("    The HTML escape sequence \"&#<d>;\" can be used to define\n")
            TEXT("    Unicode characters (<d> is a placeholder for the decimal\n")
            TEXT("    Unicode number).\n\n")
            TEXT("  MISC sleep msec\n\n")
            TEXT("    wait for specified duration\n\n")
            TEXT("  MISC testentry key value\n\n")
            TEXT("    returns 0 if a REG_MULTI_SZ registry value exists.\n\n")
            TEXT(""));
        return ERROR_INVALID_PARAMETER;
    }

    if (_tcsicmp(argv[1], TEXT("appentry")) == 0) {
        HKEY   hRootKey;
        LPTSTR pszKeyName = GetRoot(argv[2], &hRootKey);
        HKEY   hKey;
        DWORD  dwType;
        DWORD  cbData;
        LONG   lResult;
        LPTSTR pszData;
        SIZE_T cchStringLen = _tcslen(argv[4]);
        SIZE_T cchDataLen;
        if (argc != 5) goto syntax_error;
        lResult = RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ|KEY_WRITE, &hKey);
        if (lResult != ERROR_SUCCESS) {
            return lResult;
        }
        lResult = RegQueryValueEx(hKey, argv[3], NULL, &dwType, 0, &cbData);
        if (lResult != ERROR_SUCCESS) {
            RegCloseKey(hKey);
            return lResult;
        }
        if (dwType != REG_MULTI_SZ) {
            RegCloseKey(hKey);
            return ERROR_INVALID_DATA;
        }
        pszData = LocalAlloc(LMEM_FIXED|LMEM_ZEROINIT, cbData + (cchStringLen + 2) *
          sizeof(TCHAR));
        if (pszData == NULL) {
            RegCloseKey(hKey);
            return ERROR_OUTOFMEMORY;
        }
        lResult = RegQueryValueEx(hKey, argv[3], NULL, &dwType, (LPBYTE) pszData, &cbData);
        if (lResult != ERROR_SUCCESS) {
            LocalFree(pszData);
            RegCloseKey(hKey);
            return lResult;
        }
        cchDataLen = MultiStringLen(pszData);
        _tcscpy(pszData + (cchDataLen - ((cchDataLen < 3) ? cchDataLen : 1)), argv[4]);
        /* '\0\0' is already set by zero initialization */
        cbData = (DWORD) (sizeof(TCHAR) * MultiStringLen(pszData));
        lResult = RegSetValueEx(hKey, argv[3], 0, REG_MULTI_SZ, (LPCBYTE) pszData, cbData);
        LocalFree(pszData);
        RegCloseKey(hKey);
        return lResult;
    } else if (_tcsicmp(argv[1], TEXT("checkstr")) == 0) {
        if (argc == 4) {
            if (_tcsicmp(argv[2], TEXT("ascii")) == 0) {
                LPCTSTR psz;
                for (psz = argv[3]; *psz; psz++) {
                    if (*psz < (TCHAR)0x20 || *psz > (TCHAR)0x7e) {
                        return 1;
                    }
                }
                return 0;
            } else if (_tcsicmp(argv[2], TEXT("nospace")) == 0) {
                return _tcschr(argv[3], TEXT(' ')) == NULL ? 0 : 1;
            } else if (_tcsicmp(argv[2], TEXT("filename")) == 0) {
                return FileNameIsValid(argv[3]) ? 0 : 1;
            } else if (_tcsicmp(argv[2], TEXT("tcfilename")) == 0) {
                return VolumeNameIsValid(argv[3]) ? 0 : 1;
            } else {
                goto syntax_error;
            }
        } else {
            goto syntax_error;
        }
    } else if (_tcsicmp(argv[1], TEXT("checkstrlen")) == 0) {
        if (argc == 4) {
            return (int)_tcslen(argv[3]) <= _tstoi(argv[2]) ? 0 : 1;
        } else {
            goto syntax_error;
        }
    } else if (_tcsicmp(argv[1], TEXT("delentry")) == 0) {
        HKEY   hRootKey;
        LPTSTR pszKeyName = GetRoot(argv[2], &hRootKey);
        HKEY   hKey;
        DWORD  dwType;
        DWORD  cbData;
        LONG   lResult;
        LPTSTR pszData;
        if (argc != 5) goto syntax_error;
        lResult = RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_READ|KEY_WRITE, &hKey);
        if (lResult != ERROR_SUCCESS) {
            return lResult;
        }
        lResult = RegQueryValueEx(hKey, argv[3], NULL, &dwType, 0, &cbData);
        if (lResult != ERROR_SUCCESS) {
            RegCloseKey(hKey);
            return lResult;
        }
        if (dwType != REG_MULTI_SZ) {
            RegCloseKey(hKey);
            return ERROR_INVALID_DATA;
        }
        pszData = LocalAlloc(LMEM_FIXED|LMEM_ZEROINIT, cbData + 2*sizeof(TCHAR));
        if (pszData == NULL) {
            RegCloseKey(hKey);
            return ERROR_OUTOFMEMORY;
        }
        lResult = RegQueryValueEx(hKey, argv[3], NULL, &dwType, (LPBYTE) pszData, &cbData);
        if (lResult != ERROR_SUCCESS) {
            LocalFree(pszData);
            RegCloseKey(hKey);
            return lResult;
        }
        cbData = (DWORD) (sizeof(TCHAR) * RemoveString(pszData, argv[4]));
        lResult = RegSetValueEx(hKey, argv[3], 0, REG_MULTI_SZ, (LPCBYTE) pszData, cbData);
        LocalFree(pszData);
        RegCloseKey(hKey);
        return lResult;
    } else if (_tcsicmp(argv[1], TEXT("delete")) == 0) {
        HKEY   hRootKey;
        LPTSTR pszKeyName = GetRoot(argv[2], &hRootKey);
        if (argc == 3) {
            if (!RegRemoveKey(hRootKey, pszKeyName)) {
                return GetLastError();
            }
        } else if (argc == 4) {
            if (!RegRemoveValue(hRootKey, pszKeyName, argv[3])) {
                return GetLastError();
            }
        } else {
            goto syntax_error;
        }
    } else if (_tcsicmp(argv[1], TEXT("drivetype")) == 0) {
        WCHAR wszDrive[4];
        if (argc != 3) {
            goto syntax_error;
        }
        wszDrive[0] = *argv[2];
        wszDrive[1] = L':';
        wszDrive[2] = L'\\';
        wszDrive[3] = L'\0';
        return GetDriveType(wszDrive);
    } else if (_tcsicmp(argv[1], TEXT("findistr")) == 0) {
        if (argc == 4) {
            _tcslwr(argv[2]);
            _tcslwr(argv[3]);
            return (_tcsstr(argv[2], argv[3]) != NULL) ? 0 : 1;
        } else {
            goto syntax_error;
        }
    } else if (_tcsicmp(argv[1], TEXT("findstr")) == 0) {
        if (argc == 4) {
            return (_tcsstr(argv[2], argv[3]) != NULL) ? 0 : 1;
        } else {
            goto syntax_error;
        }
    } else if (_tcsicmp(argv[1], TEXT("isadmin")) == 0) {
        return IsUserAnAdmin() ? NO_ERROR : ERROR_ACCESS_DENIED;
    } else if (_tcsicmp(argv[1], TEXT("iswin64")) == 0) {
#ifndef _WIN64
        return IsWow64() ? NO_ERROR : ERROR_INVALID_PARAMETER;
#else
        return NO_ERROR;
#endif
    } else if (_tcsicmp(argv[1], TEXT("osversion")) == 0) {
        OSVERSIONINFO osvi;
        osvi.dwOSVersionInfoSize = sizeof(osvi);
        if (!GetVersionEx(&osvi) || osvi.dwPlatformId != VER_PLATFORM_WIN32_NT ||
          osvi.dwMajorVersion < 5) {
            return 1;
        }
        return osvi.dwMajorVersion * 100 + osvi.dwMinorVersion;
    } else if (_tcsicmp(argv[1], TEXT("print")) == 0) {
        if (argc != 3) {
            goto syntax_error;
        }
        _tprintf(TEXT("%s"), argv[2]);
    } else if (_tcsicmp(argv[1], TEXT("protectfile")) == 0) {
        if (argc != 3) {
            goto syntax_error;
        }
        if (!ProtectFile(argv[2],
          TEXT("D:")
          TEXT("(A;OICI;GA;;;SY)")
          TEXT("(A;OICI;GA;;;BA)")
          TEXT("(A;OICI;GRGX;;;PU)")
          TEXT("(A;OICI;GRGX;;;BU)")
          TEXT(""))) {
            dwErrorCode = GetLastError();
        }
    } else if (_tcsicmp(argv[1], TEXT("protectsystemfile")) == 0) {
        if (argc != 3) {
            goto syntax_error;
        }
        if (!ProtectFile(argv[2],
          TEXT("D:")
          TEXT("(A;OICI;GA;;;SY)")
          TEXT("(A;OICI;GA;;;BA)")
          TEXT(""))) {
            dwErrorCode = GetLastError();
        }
    } else if (_tcsicmp(argv[1], TEXT("set")) == 0) {
        HKEY   hRootKey;
        int    iType = GetType(argv[2]);
        LPTSTR pszKeyName = GetRoot(argv[3], &hRootKey);
        HKEY   hKey;
        LONG   lResult;

        if (argc != 6) goto syntax_error;

        if (iType == REG_NONE || pszKeyName == NULL) goto syntax_error;

        lResult = RegCreateKeyEx(hRootKey, pszKeyName, 0, NULL, 0,
          KEY_WRITE, NULL, &hKey, NULL);
        if (lResult != ERROR_SUCCESS) return lResult;

        if (iType == REG_SZ) {
            DWORD cbValueData;
            LPTSTR pszValueData = ConvertString(argv[5], FALSE, &cbValueData);
            if (pszValueData == NULL) {
                return ERROR_NOT_ENOUGH_MEMORY;
            }
            lResult = RegSetValueEx(hKey, argv[4], 0, REG_SZ,
              (const BYTE*)pszValueData, cbValueData);
            LocalFree(pszValueData);
            if (lResult != ERROR_SUCCESS) return lResult;
        } else if (iType == REG_MULTI_SZ) {
            DWORD cbValueData;
            LPTSTR pszValueData = ConvertString(argv[5], TRUE, &cbValueData);
            if (pszValueData == NULL) {
                return ERROR_NOT_ENOUGH_MEMORY;
            }
            lResult = RegSetValueEx(hKey, argv[4], 0, REG_MULTI_SZ,
              (const BYTE*)pszValueData, cbValueData);
            LocalFree(pszValueData);
            if (lResult != ERROR_SUCCESS) return lResult;
        } else if (iType == REG_EXPAND_SZ) {
            DWORD cbValueData;
            LPTSTR pszValueData = ConvertString(argv[5], FALSE, &cbValueData);
            if (pszValueData == NULL) {
                return ERROR_NOT_ENOUGH_MEMORY;
            }
            lResult = RegSetValueEx(hKey, argv[4], 0, REG_EXPAND_SZ,
              (const BYTE*)pszValueData, cbValueData);
            LocalFree(pszValueData);
            if (lResult != ERROR_SUCCESS) return lResult;
        } else {
            DWORD dwData = _tcstoul(argv[5], NULL, 0);
            lResult = RegSetValueEx(hKey, argv[4], 0, REG_DWORD,
              (const BYTE*)&dwData, sizeof(DWORD));
            if (lResult != ERROR_SUCCESS) return lResult;
        }
        RegCloseKey(hKey);
    } else if (_tcsicmp(argv[1], TEXT("sleep")) == 0) {
        if (argc != 3) {
            goto syntax_error;
        }
        Sleep(_tcstoul(argv[2], NULL, 0));
        return NO_ERROR;
    } else if (_tcsicmp(argv[1], TEXT("testentry")) == 0) {
        HKEY   hRootKey;
        LPTSTR pszKeyName = GetRoot(argv[2], &hRootKey);
        HKEY   hKey;
        DWORD  dwType;
        LONG   lResult;
        if (argc != 4) goto syntax_error;
        lResult = RegOpenKeyEx(hRootKey, pszKeyName, 0, KEY_QUERY_VALUE, &hKey);
        if (lResult != ERROR_SUCCESS) {
            return lResult;
        }
        lResult = RegQueryValueEx(hKey, argv[3], NULL, &dwType, 0, NULL);
        RegCloseKey(hKey);
        if (lResult != ERROR_SUCCESS) {
            return lResult;
        }
        return (dwType == REG_MULTI_SZ) ? NO_ERROR : ERROR_INVALID_DATA;
    } else {
        goto syntax_error;
    }

    if (dwErrorCode) {

        LPTSTR pszBuffer = NULL;

        _putts(TEXT("MISC V") TEXT(PROJECT_VERSION_TEXT) TEXT("\n"));

        if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
          FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ARGUMENT_ARRAY,
          NULL, dwErrorCode, LANG_NEUTRAL, (LPTSTR) &pszBuffer, 0, NULL)) {
            _putts(pszBuffer);
            LocalFree(pszBuffer);
        } else {
            _tprintf(TEXT("Error Code = %d\n"), dwErrorCode);
        }
    }

    return dwErrorCode;
}
